"""Data collection from Github."""
from time import sleep
from random import randrange
from datetime import datetime
from github import GithubException
from raven import breadcrumbs
from pygit2 import GitError
from gitalizer.helpers.config import config

from gitalizer.extensions import github, sentry, logger
from gitalizer.models.repository import Repository
from gitalizer.helpers.parallel import new_session
from gitalizer.helpers.parallel.manager import Manager
from gitalizer.helpers.parallel.messages import error_message
from gitalizer.aggregator.git.commit import CommitScanner
from gitalizer.aggregator.git.repository import get_git_repository, delete_git_repository
from gitalizer.aggregator.github import (
    call_github_function,
    get_github_object,
)


def get_github_repository_by_owner_name(owner: str, name: str):
    """Get a repository by it's owner and name."""
    full_name = f'{owner}/{name}'
    response = get_github_repository(full_name)
    logger.info(response['message'])
    if 'error' in response:
        logger.error(response['error'])


def get_github_repository_users(full_name: str):
    """Get all collaborators of a repository."""
    repo = call_github_function(github.github, 'get_repo', [full_name])
    collaborators = call_github_function(repo, 'get_collaborators')
    while collaborators._couldGrow():
        call_github_function(collaborators, '_grow')

    collaborator_list = [c.login for c in collaborators]

    sub_manager = Manager('github_repository', [])
    manager = Manager('github_contributor', collaborator_list, sub_manager)
    manager.start()
    manager.run()


def get_github_repository(full_name: str):
    """Get all information from a single repository."""
    try:
        session = new_session()
        # Sleep for a random time to avoid hitting the abuse detection.
        sleeptime = randrange(1, 15)
        sleep(sleeptime)

        github_repo = call_github_function(github.github, 'get_repo',
                                           [full_name], {'lazy': False})

        repository = Repository.get_or_create(
            session,
            github_repo.ssh_url,
            name=github_repo.name,
            full_name=github_repo.full_name,
        )

        if repository.broken:
            return {'message': f'Skip broken repo {github_repo.ssh_url}'}
        elif github_repo.size > int(config['aggregator']['max_repository_size']):
            repository.too_big = True
            session.add(repository)
            session.commit()
            sentry.captureMessage(f'Repo filesize too big', level='info',
                                  extra={'repo': repository.clone_url})

            return {'message': f'Repo too big (filesize): {github_repo.ssh_url}'}

        current_time = datetime.now().strftime('%H:%M')

        owner = get_github_object(github_repo, 'owner')
        git_repo = get_git_repository(
            github_repo.ssh_url,
            owner.login,
            github_repo.name,
        )
        scanner = CommitScanner(git_repo, session, github_repo)
        commit_count = scanner.scan_repository()

        breadcrumbs.record(
            data={'action': 'Commits scanned. Set repo metadata and debug output'},
            category='info',
        )

        repository = session.query(Repository).get(github_repo.ssh_url)
        rate = github.github.get_rate_limit().core
        time = rate.reset.strftime("%H:%M")
        current_time = datetime.now().strftime('%H:%M')

        message = f'{current_time}: '
        message += f'Scanned {repository.clone_url} with {commit_count} commits.\n'
        message += f'{rate.remaining} of 5000 remaining. Reset at {time}\n'

        response = {'message': message}

        repository.updated_at = datetime.now()
        session.add(repository)
        session.commit()

    except GithubException as e:
        # 451: Access denied. Repository probably gone private.
        # 404: User or repository just got deleted
        if e.status == 451 or e.status == 404:
            repository = session.query(Repository) \
                .filter(Repository.full_name == full_name) \
                .one_or_none()

            if repository:
                repository.broken = True
                session.add(repository)
                session.commit()
            response = {'message': 'Repository access blocked.'}
        # Catch any other GithubException
        else:
            sentry.captureException()
            response = error_message('Error in get_repository:\n')

        pass

    except (GitError, UnicodeDecodeError):
        response = error_message('Error in get_repository:\n')
        pass

    except BaseException:
        # Catch any exception and print it, as we won't get any information due to threading otherwise.
        sentry.captureException()
        response = error_message('Error in get_repository:\n')
        pass

    finally:
        if 'owner' in locals() and 'github_repo' in locals():
            delete_git_repository(owner.login, github_repo.name)
        session.close()

    return response



"""Data collection from Github."""
from collections import deque
from pygit2 import Repository, GitError
from sqlalchemy.orm import joinedload
from github import Repository as Github_Repository
from github.GithubObject import NotSet
from datetime import datetime, timedelta, timezone
from sqlalchemy.exc import (
    IntegrityError,
    OperationalError,
    InvalidRequestError,
)

from gitalizer.extensions import sentry
from gitalizer.aggregator.github import call_github_function
from gitalizer.models import (
    Email,
    Commit,
    Contributor,
    Repository as RepositoryModel,
)


class CommitScanner():
    """Get features from all commits in a repository."""

    def __init__(self, git_repo: Repository,
                 session,
                 github_repo: Github_Repository=None):
        """Initialize a new CommitChecker."""
        self.session = session
        self.repository = session.query(RepositoryModel) \
            .options(joinedload(RepositoryModel.commits_by_hash)) \
            .get(github_repo.ssh_url)
        self.git_repo = git_repo
        self.github_repo = github_repo
        self.queue = deque()
        self.scanned_commits = 0
        self.diffs = {}

        self.emails = {}

    def scan_repository(self):
        """Get all commits from this repository.

        It extracts the user as well as additions, deletions and the timestamp.
        This function gathers all commits in the git commit tree
        """
        commits_to_scan = self.get_commits_to_scan()
        if not commits_to_scan:
            self.repository.completely_scanned = True
            return 0

        existing_commits = self.preload_commits(commits_to_scan)

        # Get emails for all commits.
        emails_to_scan = self.unique_emails(commits_to_scan)
        self.preload_emails(emails_to_scan)
        self.collect_emails(emails_to_scan)
        self.session.commit()

        # Actually scan the commits
        for commit in commits_to_scan:
            self.scan_commit(commit, existing_commits)
            # Commit session every 20 commits to avoid loss of all data on crash.
            self.scanned_commits += 1
            if self.scanned_commits % 1000 == 0:
                self.session.commit()

        self.session.commit()

        # Set the time of the first commit as repository creation time if it isn't set yet.
        self.repository.completely_scanned = True
        if not self.repository.created_at:
            timestamp = commits_to_scan[-1].author.time
            utc_offset = timezone(timedelta(minutes=commits_to_scan[-1].author.offset))
            self.repository.created_at = datetime.fromtimestamp(timestamp, utc_offset)

        self.session.add(self.repository)
        self.session.commit()
        return self.scanned_commits

    def get_commits_to_scan(self):
        """Walk through the repository and get all commits reachable in master."""
        try:
            master_commit = self.git_repo.head.get_object()
            self.queue.appendleft(master_commit)
        except GitError as e:
            sentry.captureException(
                extra={
                    'message': 'GitError during repo cloning. Probably empty',
                    'clone_url': self.repository.clone_url,
                },
            )
            return self.scanned_commits

        # List of commit hashes to check if we already were at this point in the tree.
        commits_to_scan = []
        known_commit_hashes = set()
        # This is a simple BFS through the git commit tree.
        # If we already know a node or already scanned a node, we don't add the parents.
        while len(self.queue) > 0:
            commit = self.queue.pop()
            # Break if we already visited this tree node
            if commit.hex in known_commit_hashes:
                continue
            known_commit_hashes.add(commit.hex)

            commit_known = commit.hex in self.repository.commits_by_hash
            # Repo has been completely scanned and a this is a known commit.
            if commit_known and self.repository.completely_scanned:
                break

            # Repo has been partially scanned and a this is a known commit.
            elif commit_known and not self.repository.completely_scanned:
                [self.queue.appendleft(parent) for parent in commit.parents]
                continue
            # This is an unknown commit.
            elif not commit_known:
                [self.queue.appendleft(parent) for parent in commit.parents]

            if len(commit.parents) == 1:
                self.diffs[commit.hex] = commit.tree.diff_to_tree(commit.parents[0].tree)
            commits_to_scan.append(commit)

            if len(commits_to_scan) > 100000:
                sentry.captureMessage(
                    'Repository too big',
                    extra={'url': self.repository.clone_url},
                    level='info',
                    tags={'type': 'too_big', 'entity': 'repository'},
                )
                self.repository.too_big = True
                self.session.add(self.repository)
                commits_to_scan = []
                break

        return commits_to_scan

    def preload_commits(self, commits_to_scan):
        """Fetch all commits from db with matching shas.

        We do this once to bunch-fetch all matching commits to avoid performance breaks.
        """
        hashes_to_scan = [c.hex for c in commits_to_scan]
        if len(hashes_to_scan) != 0:
            existing_commits = self.session.query(Commit) \
                .filter(Commit.sha.in_(hashes_to_scan)) \
                .all()
            existing_commits = {c.sha: c for c in existing_commits}
        else:
            existing_commits = {}
        return existing_commits

    def scan_commit(self, git_commit, existing_commits):
        """Get all features of a specific commit."""
        # If we already know this commit just add the commit to this repository.
        if git_commit.hex in existing_commits:
            commit = existing_commits[git_commit.hex]
            commit.repositories.append(self.repository)
            self.session.add(commit)

        # Unknown commit, thereby we need to get all information
        else:
            try:
                author_email = self.emails[git_commit.author.email]
                committer_email = self.emails[git_commit.committer.email]
                commit = Commit(git_commit.hex, self.repository,
                                author_email, committer_email)

                diff = self.diffs.get(git_commit.hex)
                if diff:
                    commit.additions = diff.stats.insertions
                    commit.deletions = diff.stats.deletions

                if git_commit.author:
                    timestamp = git_commit.author.time
                    offset_delta = timedelta(minutes=git_commit.author.offset)
                    utc_offset = timezone(offset_delta)
                    commit.creation_time = datetime.fromtimestamp(timestamp, utc_offset)

                timestamp = git_commit.commit_time
                offset_delta = timedelta(minutes=git_commit.commit_time_offset)
                utc_offset = timezone(offset_delta)
                commit.commit_time = datetime.fromtimestamp(timestamp, utc_offset)
                commit.commit_time_offset = offset_delta

                self.session.add(commit)
            except BaseException as e:
                sentry.captureException(
                    extra={
                        'message': 'Error during Commit creation',
                        'clone_url': self.repository.clone_url,
                        'hex': git_commit.hex,
                    },
                )

    def unique_emails(self, commits):
        """Get all commits that should be collected."""
        checked_emails = set()
        emails_to_scan = []
        for commit in commits:
            if commit.author.email not in checked_emails:
                emails_to_scan.append((commit, 'author', commit.author.email))
                checked_emails.add(commit.author.email)

            if commit.committer.email not in checked_emails:
                emails_to_scan.append((commit, 'committer', commit.committer.email))
                checked_emails.add(commit.committer.email)

        return emails_to_scan

    def preload_emails(self, emails_to_scan):
        """Bulk preload all emails to avoid DB overhead."""
        addresses = [e[2] for e in emails_to_scan]
        emails = self.session.query(Email) \
            .filter(Email.email.in_(addresses)) \
            .all()
        if emails:
            self.emails = {e.email: e for e in emails}
        else:
            self.emails = {}

    def collect_emails(self, emails):
        """Get emails of all commits to scan."""
        collecting = True
        one_more_try = False
        while collecting:
            try:
                email_count = 0
                for (commit, address_type, address) in emails:
                    email = self.emails.get(address)

                    # Create a new mail if we don't know it yet
                    if not email:
                        email = Email.get_email(
                            address,
                            self.session,
                            do_commit=False,
                        )
                        self.emails[address] = email

                    # Get the email relation to github contributors
                    if address_type == 'author':
                        self.get_github_author(email, commit, do_commit=False)
                    if address_type == 'committer':
                        self.get_github_committer(email, commit, do_commit=False)

                    # Add the contributor to the repository if he isn't known yet.
                    if email.contributor:
                        if self.repository not in email.contributor.repositories:
                            email.contributor.repositories.append(self.repository)
                    else:
                        # Mark email as unknown to prevent further github queries for this email.
                        email.unknown = True
                    self.session.add(email)
                    email_count += 1

                    # Commit all 20 emails to prevent large reloads on transaction failures.
                    if email_count % 20:
                        self.session.commit()
                        one_more_try = False
                        collecting = True

                collecting = False
            except (IntegrityError, OperationalError, InvalidRequestError) as e:
                # This can happen in case two threads add the same email addresses
                # Rollback the transaction, try one more time and raise if it fails again.
                # If the next transaction is successful, we reset the retry behaviour.
                sentry.captureMessage(f'Email collection IntegrityError', level='info')
                self.session.rollback()
                self.preload_emails(emails)
                if one_more_try:
                    collecting = False
                    raise e

                one_more_try = True

    def get_github_author(self, email, git_commit, do_commit=True):
        """Get the related Github author."""
        # No Github repository or the contributor is already known
        if not self.github_repo or email.contributor is not None or email.unknown:
            return
        github_commit = call_github_function(self.github_repo, 'get_commit', [git_commit.hex])

        if github_commit.author and github_commit.author is not NotSet:
            # Workaround for issue https://github.com/PyGithub/PyGithub/issues/279
            if github_commit.author._url.value is None:
                sentry.captureMessage('Author has no _url', level='info')
                return

            contributor = Contributor.get_contributor(
                github_commit.author.login,
                self.session,
                do_commit=do_commit,
            )
            email.contributor = contributor

    def get_github_committer(self, email, git_commit, do_commit=True):
        """Get the related Github committer."""
        # No Github repository or the contributor is already known
        if not self.github_repo or email.contributor is not None or email.unknown:
            return
        github_commit = call_github_function(self.github_repo, 'get_commit', [git_commit.hex])

        if github_commit.committer and github_commit.committer is not NotSet:
            # Workaround for issue https://github.com/PyGithub/PyGithub/issues/279
            if github_commit.committer._url.value is None:
                sentry.captureMessage('committer has no _url', level='info')
                return

            contributor = Contributor.get_contributor(
                github_commit.committer.login,
                self.session,
                do_commit=do_commit,
            )
            email.contributor = contributor
        return



"""Helper functions for github api calls."""
from github.GithubException import RateLimitExceededException, GithubException

import time
from socket import timeout
from random import randrange
from raven import breadcrumbs
from datetime import datetime, timedelta

from gitalizer.extensions import github, logger


def call_github_function(github_object: object, function_name: str,
                         args: list = None, kwargs: dict = None):
    """Call a pygithub object member function.

    We need to handle those calls in case we get rate limited.
    """
    _try = 0
    tries = 5
    exception = None
    while _try <= tries:
        try:
            if not args:
                args = []
            if not kwargs:
                kwargs = {}
            retrieved_object = getattr(github_object, function_name)(*args, **kwargs)
            return retrieved_object
        except RateLimitExceededException as e:
            # Wait until the rate limiting is reset
            resettime = github.github.get_rate_limit().core.reset
            if resettime < datetime.now():
                resettime = datetime.now()
            delta = resettime - datetime.utcnow()
            delta += timedelta(minutes=2)
            total_minutes = int(delta.total_seconds() / 60)
            logger.info('Hit the rate limit.')
            logger.info(f'Reset at {resettime}. Waiting for {total_minutes} minutes.')
            time.sleep(delta.total_seconds())

            _try += 1
            exception = e
            pass
        except GithubException as e:
            # Forbidden or not found (Just made private or deleted)
            if e.status == 451 or e.status == 404:
                raise e

                # Otherwise abuse detection
            if e.status == 403:
                seconds = randrange(180, 480)
                logger.info('Github abuse detection.')
                logger.info(f'Waiting for {seconds} seconds')
                time.sleep(seconds)

            breadcrumbs.record(
                data={'action': 'Github Exception.', 'exception': e},
                category='info',
            )

            _try += 1
            exception = e
            pass
        except timeout as e:
            logger.info('Hit socket timeout waiting 10 secs.')
            breadcrumbs.record(data={'action': 'Socket timeout hit'},
                               category='info')

            time.sleep(10)
            _try += 1
            exception = e
            pass

    raise exception


def get_github_object(github_object: object, object_name: str):
    """Get a pygithub object.

    As pygithub sometimes implicitly queries the github api on a class member access,
    we need to handle those accesses in case we get rate limited.
    """
    _try = 0
    tries = 5
    exception = None
    while _try <= tries:
        try:
            retrieved_object = getattr(github_object, object_name)
            return retrieved_object
        except RateLimitExceededException as e:
            # Wait until the rate limiting is reset
            resettime = github.github.get_rate_limit().core.reset
            if resettime < datetime.now():
                resettime = datetime.now()
            delta = resettime - datetime.now()
            delta += timedelta(minutes=2)
            total_minutes = int(delta.total_seconds() / 60)
            logger.info('Hit the rate limit.')
            logger.info(f'Reset at {resettime}. Waiting for {total_minutes} minutes.')

            time.sleep(delta.total_seconds())

            _try += 1
            exception = e
            pass
        except GithubException as e:
            # Forbidden or not found (Just made private or deleted)
            if e.status == 451 or e.status == 404:
                raise e

                # Otherwise abuse detection
            if e.status == 403:
                seconds = randrange(180, 480)
                logger.info('Github abuse detection.')
                logger.info(f'Waiting for {seconds} seconds')
                time.sleep(seconds)

            breadcrumbs.record(
                data={'action': 'Github Exception.', 'exception': e},
                category='info',
            )

            _try += 1
            exception = e
            pass
        except timeout as e:
            logger.info('Hit socket timeout waiting 10 secs.')
            breadcrumbs.record(data={'action': 'Socket timeout hit'},
                               category='info')

            time.sleep(10)
            _try += 1
            exception = e
            pass

    raise exception



"""Clone repositories and get data from it."""
import os
import shutil
import pygit2
from pygit2 import Repository, clone_repository

from gitalizer.helpers.config import config


def get_git_repository(url: str, owner: str, name: str):
    """Clone or update a repository."""
    base_dir = config['cloning']['temporary_clone_path']
    clone_dir = os.path.join(base_dir, owner, name)

    if os.path.exists(clone_dir):
        shutil.rmtree(clone_dir)

    callbacks = None
    if 'https://' not in url:
        keypair = pygit2.Keypair(username=config['cloning']['ssh_user'],
                                 pubkey=config['cloning']['public_key'],
                                 privkey=config['cloning']['private_key'],
                                 passphrase=config['cloning']['ssh_password'])
        callbacks = pygit2.RemoteCallbacks(credentials=keypair)

    os.makedirs(clone_dir)
    repo = clone_repository(url, clone_dir, bare=True, callbacks=callbacks)
    repo = Repository(clone_dir)

    return repo


def delete_git_repository(owner: str, name: str):
    """Delete a git repository."""
    base_dir = config['cloning']['temporary_clone_path']
    clone_dir = os.path.join(base_dir, owner, name)
    if os.path.exists(clone_dir):
        shutil.rmtree(clone_dir)



"""Data collection from Github."""
import traceback
from datetime import datetime
from github.GithubException import GithubException

from gitalizer.helpers.config import config
from gitalizer.models import Repository, Contributor
from gitalizer.extensions import github, sentry, db, logger
from gitalizer.aggregator.github import call_github_function, get_github_object
from gitalizer.helpers.parallel import new_session
from gitalizer.helpers.parallel.manager import Manager
from gitalizer.helpers.parallel.messages import (
    user_too_big_message,
    user_up_to_date_message,
)


def get_user_with_followers(name: str):
    """Get all relevant Information about all friends of a specific user.."""
    user = call_github_function(github.github, 'get_user', [name])
    followers = call_github_function(user, 'get_followers')
    following = call_github_function(user, 'get_following')

    # Add all following and followed people into list
    # Deduplicate the list as we have to make as few API calls as possible.
    user_list = [user]
    for follower in followers:
        user_list.append(follower)
    for followed in following:
        exists = filter(lambda x: x.login == followed.login, user_list)
        if len(list(exists)) == 0:
            user_list.append(followed)

    user_logins = [u.login for u in user_list]
#    for user in user_list:
#        print(user)
    sub_manager = Manager('github_repository', [])
    manager = Manager('github_contributor', user_logins, sub_manager)
    manager.start()
    manager.run()

    try:
        session = new_session()
        for login in user_logins:
            contributor = session.query(Contributor) \
                .filter(Contributor.login.ilike(login)) \
                .one()
            if not contributor.too_big:
                contributor.last_full_scan = datetime.utcnow()
                session.add(contributor)
        session.commit()
    finally:
        session.close()


def get_user(login: str):
    """Get a user by his login name."""
    user = call_github_function(github.github, 'get_user', [login])
    sub_manager = Manager('github_repository', [])
    manager = Manager('github_contributor', [user.login], sub_manager)
    manager.start()
    manager.run()

    try:
        session = db.get_session()
        contributor = session.query(Contributor) \
            .filter(Contributor.login.ilike(login)) \
            .one()
        contributor.last_full_scan = datetime.utcnow()
        session.add(contributor)
        session.commit()
    finally:
        session.close()


def get_user_repos(user_login: str, skip=True):
    """Get all relevant Information for a single user."""
    try:
        session = new_session()
        contributor = Contributor.get_contributor(user_login, session, True)
        # Checks for already scanned users.
        if not contributor.should_scan():
            return user_up_to_date_message(user_login)
        if contributor.too_big:
            return user_too_big_message(user_login)

        user = call_github_function(github.github, 'get_user', [user_login])
        owned = user.get_repos()
        starred = user.get_starred()
        repos_to_scan = set()

        # Prefetch all owned repositories
        user_too_big = False
        owned_repos = 0
        while owned._couldGrow() and not user_too_big:
            owned_repos += 1
            call_github_function(owned, '_grow')

            # Debug messages to see that the repositories are still collected.
            if owned_repos % 100 == 0:
                logger.info(f'{owned_repos} owned repos for user {user_login}.')

            # The user is too big. Just drop him.
            if skip and owned_repos > int(config['aggregator']['max_repositories_for_user']):
                user_too_big = True

        # Prefetch all starred repositories
        starred_repos = 0
        while starred._couldGrow() and not user_too_big:
            starred_repos += 1
            call_github_function(starred, '_grow')
            # Debug messages to see that the repositories are still collected.
            if starred_repos % 100 == 0:
                logger.info(f'{starred_repos} starred repos for user {user_login}.')

            # The user is too big. Just drop him.
            if skip and starred_repos > int(config['aggregator']['max_repositories_for_user']):
                user_too_big = True

        # User has too many repositories. Flag him and return
        if user_too_big:
            contributor.too_big = True
            sentry.captureMessage(
                'User too big',
                extra={'url': contributor.login},
                level='info',
                tags={'type': 'too_big', 'entity': 'user'},
            )
            session.add(contributor)
            session.commit()
            return user_too_big_message(user_login)

        # Check own repositories. We assume that we are collaborating in those
        for github_repo in owned:
            repository = Repository.get_or_create(
                session,
                github_repo.ssh_url,
                name=github_repo.name,
                full_name=github_repo.full_name,
            )
            if github_repo.fork and not repository.is_invalid():
                check_fork(github_repo, session, repository,
                           repos_to_scan, user_login)
            session.add(repository)

            if not repository.should_scan():
                continue

            session.commit()
            repos_to_scan.add(github_repo.full_name)

        # Check stars and if the user collaborated to them.
        for github_repo in starred:
            repository = Repository.get_or_create(
                session,
                github_repo.ssh_url,
                name=github_repo.name,
                full_name=github_repo.full_name,
            )

            if github_repo.fork and not repository.is_invalid():
                check_fork(github_repo, session, repository,
                           repos_to_scan, user_login)
            session.add(repository)

            if not repository.should_scan():
                continue

            repos_to_scan.add(github_repo.full_name)

        session.commit()

        rate = github.github.get_rate_limit().core
        message = f'Got repositories for {user.login}. '
        message += f'{user.login}. {rate.remaining} of 5000 remaining.'
        response = {
            'message': message,
            'tasks': list(repos_to_scan),
        }
    except BaseException:
        # Catch any exception and print it, as we won't get any information due to threading otherwise.
        sentry.captureException()
        response = {
            'message': f'Error while getting repos for {user_login}:\n',
            'error': traceback.format_exc(),
        }
        pass
    finally:
        session.close()

    return response


def get_user_data(user_data: tuple):
    """Get all missing data from a user."""
    try:
        contributor = user_data[0]
        login = contributor.login

        session = new_session()
        contributor = Contributor.get_contributor(login, session, True)

        user = call_github_function(github.github, 'get_user', [login])
        if user.location:
            contributor.location = user.location

        session.add(contributor)
        session.commit()
        response = {'message': f'Scanned user {login}'}

    except GithubException as e:
        # Forbidden or not found (Just made private or deleted)
        if e.status == 404:
            response = {'message': f'User {login} not found.'}
        pass

    except BaseException as e:
        # Catch any exception and print it, as we won't get any information due to threading otherwise.
        sentry.captureException()
        response = {
            'message': f'Error while getting repos for {login}:\n',
            'error': traceback.format_exc(),
        }
        pass
    finally:
        session.close()

    return response


def check_fork(github_repo, session, repository, scan_list, user_login=None):
    """Handle github_repo forks."""
    # We already scanned this repository and only need to check
    # if it or its parent should be scanned
    if repository.completely_scanned:
        # Its a fork, check if the parent needs to be scanned
        if repository.fork:
            if repository.parent.should_scan():
                scan_list.add(github_repo.parent.full_name)
        # Its no fork just skip and return
        else:
            return

    # We don't know the repository yet.
    # Create the parent and check if it is a valid fork
    try:
        get_github_object(github_repo, 'parent')
        call_github_function(github_repo.parent, '_completeIfNeeded')
    except GithubException as e:
        if e.status == 451 or e.status == 404:
            repository.fork = False
            return

    parent_repository = Repository.get_or_create(
        session,
        github_repo.parent.ssh_url,
        name=github_repo.parent.name,
        full_name=github_repo.parent.full_name,
    )

    # If the names are identical it's likely not spite/hate fork.
    if github_repo.parent.name == github_repo.name:
        # Set the parent on the forked repository
        if not repository.parent:
            repository.parent = parent_repository

        # Mark the repository as a fork and scan the parent.
        repository.fork = True
        if parent_repository.should_scan():
            scan_list.add(parent_repository.full_name)

    session.add(repository)



"""Configuration Reader."""
import os
import sys
import configparser


def read_config():
    """Read a previous configuration file or create a new with default values."""
    path = os.path.expanduser('~/.config/gitalizer.ini')
    path = os.path.realpath(path)
    config = configparser.ConfigParser()
    # Try to get configuration file and return it
    # If this doesn't work, a new default config file will be created
    if os.path.exists(path):
        try:
            config.read(path)
            return config
        except Exception:
            print('Error while parsing config file')
            raise Exception

    config['develop'] = {
        'log_dir': './logs',
        'sentry_token': '',
        'sentry_enabled': 'no',
    }
    config['github'] = {
        'github_user': 'git',
        'github_password': '',
        'github_token': '',
    }
    config['cloning'] = {
        'ssh_user': '',
        'ssh_password': '',
        'private_key': '/home/user/.ssh/id_rsa',
        'public_key': '/home/user/.ssh/id_rsa.pub',
        'temporary_clone_path': '/tmp/gitalizer',
    }
    config['aggregator'] = {
        'git_user_scan_threads': 4,
        'git_commit_scan_threads': 4,
        'max_repository_size': 4 * 1024 * 1024,
        'max_repositories_for_user': 3000,
        'repository_rescan_interval': 21 * 24 * 60 * 60,
        'contributor_rescan_interval': 22 * 24 * 60 * 60,
    }

    config['database'] = {
        'uri': 'postgres://localhost/gitalizer',
        'echo': 'no',
    }
    config['plotting'] = {
        'plot_dir': './plots',
    }

    with open(path, 'w') as fd:
        config.write(fd)

    print('Initialized empty configuration. Please adjust before proceeding.')
    sys.exit(0)


config = read_config()



"""Data collection from Github."""

import pytz
from datetime import datetime, timedelta

from gitalizer.extensions import github, logger
from gitalizer.models import Contributor, Organization, Repository
from gitalizer.aggregator.github import call_github_function
from gitalizer.helpers.parallel import new_session
from gitalizer.helpers.parallel.manager import Manager
from gitalizer.aggregator.github.user import check_fork


def get_organization_memberships():
    """Refresh all user organizations."""
    session = new_session()

    tz = pytz.timezone('Europe/Berlin')
    now = datetime.now(tz)
    contributors = session.query(Contributor).all()
    for contributor in contributors:
        if contributor.last_full_scan and contributor.last_full_scan > now - timedelta(days=2):
            continue
        logger.info(f'Checking {contributor.login}. {github.github.rate_limiting[0]} remaining.')

        github_user = call_github_function(github.github, 'get_user',
                                           [contributor.login])

        github_orgs = call_github_function(github_user, 'get_orgs')
        for org in github_orgs:
            organization = Organization.get_organization(org.login, org.url, session)
            contributor.organizations.append(organization)
        contributor.last_full_scan = datetime.utcnow()
        session.add(contributor)
        session.commit()


def get_github_organization(name: str, members=False):
    """Get all collaborators of an organization."""
    session = new_session()
    orga = call_github_function(github.github, 'get_organization', [name])

    # Get orga repos
    orga_repos = call_github_function(orga, 'get_repos')
    while orga_repos ._couldGrow():
        call_github_function(orga_repos, '_grow')

    # Check orga repos
    repos_to_scan = set()
    for github_repo in orga_repos:
        repository = Repository.get_or_create(
            session,
            github_repo.ssh_url,
            name=github_repo.name,
            full_name=github_repo.full_name,
        )
        if github_repo.fork:
            check_fork(github_repo, session, repository, repos_to_scan)
        session.add(repository)

        if not repository.should_scan():
            continue

        session.commit()
        repos_to_scan.add(github_repo.full_name)

    member_list = set()
    if members:
        # Get members
        members = call_github_function(orga, 'get_members')
        while members._couldGrow():
            call_github_function(members, '_grow')
        member_list = set([m.login for m in members])

    # Create and start manager with orga repos and memeber_list
    sub_manager = Manager('github_repository', repos_to_scan)
    manager = Manager('github_contributor', member_list, sub_manager)
    manager.start()
    manager.run()
